module Yorinter
  class Engine
    def initialize document, params = {}, options = {}
      @document = document
      @params = params
      @engine = build_engine
      @options = options
    end

    def document_model
      @document.model || {}
    end

    def model_endpoint
      if @document.model_endpoint
        PrintableDocument.new(
          template: @document.model_endpoint
        ).render ErbEngine, @params
      end
    end

    # Get model from endpoint
    def endpoint_model
      if endpoint = model_endpoint
        JSON.parse Net::HTTP.get(URI(endpoint))
      else
        {}
      end
    end

    def model_binding_object
      model = document_model.deep_merge(endpoint_model)
      JSON.parse(model.to_json, object_class: OpenStruct)
    end

    def params_binding_object
      JSON.parse((@params || {}).to_json, object_class: OpenStruct)
    end

    def binding_object
      return @binding_object if @binding_object

      # Use document binding_object if it already has
      if @document.respond_to?(:binding_object) and @document.binding_object
        return @binding_object = @document.binding_object
      end

      this = self
      helper = @options[:helper]

      helper_wrapper = Object.new
      helper_wrapper.instance_eval @document.helper.to_s

      @binding_object = Object.new.instance_eval do
        @model  = this.model_binding_object
        @params = this.params_binding_object

        # Helper from params
        extend helper if helper

        # Helper from document
        helper_wrapper.singleton_class.constants.each do |const|
          extend helper_wrapper.singleton_class.const_get const
        end

        def get_binding
          binding
        end

        self
      end
    end

    def document_options
      return @document_options if @document_options

      @document_options = (@document.options || {}).deep_symbolize_keys

      # Render title
      if @document_options[:title]
        @document_options[:title] = render_partial @document_options[:title]
      end

      # Render footer content
      if footer = @document_options[:footer]
        if footer[:content]
          footer[:content] = render_partial footer[:content]
        end
      end

      @document_options
    end

    def render_partial template
      PrintableDocument.new(
        template: template,
        binding_object: binding_object
      ).render ErbEngine, @params
    end

    # Abstract methods
    def build_engine
      raise 'Abstract method'
    end

    def render
      raise 'Abstract method'
    end

    class << self
      def search_engine input, output
        (@@engines || {})[[input, output]]
      end

      def register_engine input, output
        @@engines ||= {}
        @@engines[[input, output]] = self
      end
    end
  end
end
